"""This module handles antenna field control via GRPC"""

import logging

from tango import DeviceProxy, DevState


from lofar_sid.interface.stationcontrol import antenna_pb2_grpc

from lofar_sid.interface.stationcontrol.antenna_pb2 import (
    AntennaReply,
    AntennaResult,
    GetAntennaRequest,
    Identifier,
    SetAntennaStatusRequest,
    SetAntennaUseRequest,
)

from tangostationcontrol.rpc.common import (
    call_exception_metrics,
    reply_on_exception,
)
from tangostationcontrol.rpc.proxy.antennadeviceproxyfactory import (
    AntennaDeviceProxyFactory,
)

logger = logging.getLogger()


class AntennaNotFoundException(ValueError):
    """Exception raised when an antenna is not found in the device."""


class Antenna(antenna_pb2_grpc.AntennaServicer):
    """Represents an antenna field in the TangoStationControl system.

    This class provides methods to control and manage antennas remotely.
    """

    def _get_antenna_index(
        self, identifier: Identifier, antennafield_device: DeviceProxy
    ) -> int:
        """Returns the current antenna index."""
        try:
            return antennafield_device.Antenna_Names_R.index(identifier.antenna_name)
        except ValueError as exc:
            logger.warning(
                "Antenna not found",
                extra={
                    "antenna_name": identifier.antenna_name,
                    "antennafield_name": identifier.antennafield_name,
                },
            )
            raise AntennaNotFoundException(
                f"Antenna '{identifier.antenna_name}' "
                f"not found in device '{identifier.antennafield_name}'."
            ) from exc

    def _get_antenna_reply(
        self,
        identifier: Identifier,
        antenna_index: int,
        antennafield_device: DeviceProxy,
    ):
        """Returns the  Message for  the caller of the GRPC."""
        try:
            antenna_status = antennafield_device.Antenna_Status_R[antenna_index]
            antenna_use = antennafield_device.Antenna_Use_R[antenna_index]
        except (AttributeError, IndexError) as ex:
            raise ValueError(
                f"Could not access device attribute for antenna "
                f"{identifier.antenna_name=} "
                f"{identifier.antenna_index}"
            ) from ex

        return AntennaReply(
            success=True,
            exception="",
            result=AntennaResult(
                identifier=identifier,
                antenna_use=antenna_use,
                antenna_status=antenna_status,
            ),
        )

    @reply_on_exception(AntennaReply)
    @call_exception_metrics("AntennaField")
    def GetAntenna(self, request: GetAntennaRequest, context):
        """Gets Correct Device, then returns the Message for  the caller of the GRPC."""
        antennafield_device = (
            AntennaDeviceProxyFactory.create_device_proxy_for_antennafield(
                request.identifier.antennafield_name
            )
        )
        antenna_index = self._get_antenna_index(request.identifier, antennafield_device)

        return self._get_antenna_reply(
            request.identifier, antenna_index, antennafield_device
        )

    def _apply_changes(self, antenna_field):
        """Apply Changes to the System. It is required to bring it off first"""

        # If the antenna_field is ON, we need to reload the changes
        if antenna_field.state() == DevState.ON:
            antenna_field.Off()
            antenna_field.Initialise()
            antenna_field.On()

            # If the antenna_field is ON, the antennas should also receive power
            # as per the station transitions.
            antenna_field.power_hardware_on()

    def _update_antenna_property(
        self,
        identifier: Identifier,
        attribute_name: str,
        property_name: str,
        new_value: int,
    ) -> AntennaReply:
        """Update the antenna Property"""
        antennafield_device = (
            AntennaDeviceProxyFactory.create_device_proxy_for_antennafield(
                identifier.antennafield_name, True
            )
        )
        antenna_index = self._get_antenna_index(identifier, antennafield_device)

        property_value = list(antennafield_device.read_attribute(attribute_name).value)
        property_value[antenna_index] = new_value

        antennafield_device.put_property({property_name: property_value})
        self._apply_changes(antennafield_device)

        return self._get_antenna_reply(identifier, antenna_index, antennafield_device)

    @reply_on_exception(AntennaReply)
    @call_exception_metrics("AntennaField")
    def SetAntennaStatus(self, request: SetAntennaStatusRequest, context):
        """Set the new Antenna  Status"""
        return self._update_antenna_property(
            request.identifier,
            "Antenna_Status_R",
            "Antenna_Status",
            request.antenna_status,
        )

    @reply_on_exception(AntennaReply)
    @call_exception_metrics("AntennaField")
    def SetAntennaUse(self, request: SetAntennaUseRequest, context):
        """Set the new Antenna  Use"""
        return self._update_antenna_property(
            request.identifier, "Antenna_Use_R", "Antenna_Use", request.antenna_use
        )
